#include "main.h"

int main(int argc, char *argv[])
{
	G::Queue<char *> qPostfixTokens;
	char *szTokens = new char[256];
	char *szToken;
	bool bRun = true;
	float fResult;

	while (bRun)
	{
		cout <<"Expression ('end' to quit): ";
		cin.getline(szTokens, 255);
		
		if (strcmp(szTokens, "end") == 0)
			break;
		
		if (Postfix::IsWhiteSpace(szTokens))
			continue;

		if (!Postfix::CreateTokenList(&szTokens))
		{
			cout <<"Error in creating token list" <<endl;
			getch();
			return 0;
		}
		
		cout <<"Token list: " <<szTokens <<endl;

		if (!Postfix::ConvertToPostfix(szTokens, qPostfixTokens))
		{
			cout <<"There was an error!" <<endl;
			getch();
			return 0;
		}
		G::Queue<char *> qTokens2(qPostfixTokens);
		cout <<"Postfix: ";
		while (qPostfixTokens.size())
		{
			qPostfixTokens.take(szToken);
			cout <<szToken <<" ";
		}
		cout <<endl;
		if (!Postfix::EvaluatePostfix(qTokens2, fResult))
			cout <<"There was an error evaluating the postfix expression" <<endl;
		else
		{
			cout <<"Result: " <<fResult <<endl;
			cout <<"Complete!" <<endl;
		}
		delete[] szTokens;
		szTokens = new char[256];
	}
	return 0;
}

bool Postfix::ConvertToPostfix(char *szTokens, G::Queue<char *> &qTokens)
{
	//G::Queue<char *> qTokens;
	G::Stack<char *> sOps;

	char *szToken = NULL;
	char *szNew = NULL;
	char *szLastToken = NULL;
	szToken = strtok(szTokens, " ");
	
	while (szToken != NULL)
	{
		//switch (szToken)
		//{
		if (strcmp(szToken, "(") == 0)
		{
			// New code for implicit *
			if (szLastToken)
			{
				int iType = Postfix::GetTokenType(szLastToken);			
				if ((!iType) || (iType == 1) || (strcmp(szLastToken,")") == 0))
				{
					char *szNewOp = NULL;
					G::Str::Copy(&szNewOp, "*");
					sOps.push(szNewOp);
				}
			}

			sOps.push(szToken);
		}
		else if (strcmp(szToken, ")") == 0)
		{
			// Loop through until ( is reached
			bool bFound = false;
			char *szOp = NULL;
			while (sOps.size() && !bFound)
			{
				sOps.pop(szOp);
				if (strcmp(szOp, "(") == 0)
					bFound = true;
				else
				{
					G::Str::Copy(&szNew, szOp);
					qTokens.give(szNew);
					szNew = NULL;
				}
			}
			if (!bFound)
			{
				// Parse error
				cout <<"Parse error, missing '('" <<endl;
				return false;
			}
		}
		else
		{
			// is it number or an operator, or variable
			unsigned int iLen = (unsigned int)strlen(szToken);
			bool bIsNumeric = true;
			for (unsigned int x = 0; x < iLen; x++)
			{
				if (!isdigit(szToken[x]) && (szToken[x] != '.') && (szToken[x] != '_'))
					bIsNumeric = false;
				if (szToken[x] == '_')
					szToken[x] = '-';
			}
			
			if (bIsNumeric)
			{
				G::Str::Copy(&szNew, szToken);
				qTokens.give(szNew);
				// Added for implicit *
				if (szLastToken)
				{
					if (strcmp(szLastToken,")") == 0)
					{
						char *szNewOp = NULL;
						G::Str::Copy(&szNewOp, "*");
						sOps.push(szNewOp);
					}
				}

				szNew = NULL;
			}
			
			else if (isalpha(szToken[0]))
			{
				// Its a variable
				G::Str::Copy(&szNew, szToken);
				qTokens.give(szNew);
				// Added for implicit *
				if (szLastToken)
				{
					if (strcmp(szLastToken,")") == 0)
					{
						char *szNewOp = NULL;
						G::Str::Copy(&szNewOp, "*");
						sOps.push(szNewOp);
					}
				}
				szNew = NULL;
			}
					
			else
			{
				// its a operator, have to get precedence
				if (!sOps.size())
					sOps.push(szToken);
				else
				{
					char *szLastOp = NULL;
					sOps.pop(szLastOp);
					if (strcmp(szLastOp, "(") == 0)
					{
						sOps.push(szLastOp);
						sOps.push(szToken);
					}
					else 
					{
						// look up precedence
						char cPrecedence = Postfix::GetOpPrecedence(szToken[0], szLastOp[0]);
						if (!cPrecedence || (cPrecedence == 1))
						{
							G::Str::Copy(&szNew, szLastOp);
							qTokens.give(szNew);
							szNew = NULL;
							bool bEnd = false;
							while (sOps.size() && !bEnd)
							{							
								char *szCow = NULL;
								sOps.pop(szCow);
								if ((strcmp(szCow, "(") == 0) || (Postfix::GetOpPrecedence(szToken[0], szCow[0]) == 2))
								{
									sOps.push(szCow);
									bEnd = true;
								}
								else
								{
									G::Str::Copy(&szNew, szCow);
									qTokens.give(szNew);
									szNew = NULL;
								}
							}
							sOps.push(szToken);
						}
						else
						{
							sOps.push(szLastOp);
							sOps.push(szToken);
						}
					}
				}
			}
		}
		szLastToken = szToken;
		szToken = strtok(NULL, " ");
	}

	// Give remaining ops to queue
	while (sOps.size())
	{
		char *szOp = NULL;
		sOps.pop(szOp);
		if (strcmp(szOp, "(") == 0)
		{
			cout <<"Parse error, missing ')'" <<endl;
			return false;
		}
		else
		{
			G::Str::Copy(&szNew, szOp);
			qTokens.give(szNew);
			szNew = NULL;
		}
	}
	//qPostfixTokens = qTokens;
	return true;
}

char Postfix::GetOpPrecedence(char cOp1, char cOp2)
{
	char cOps[2];
	cOps[0] = cOp1;
	cOps[1] = cOp2;
	char cP;
	for (int x = 0; x < 2; x++)
	{
		switch (cOps[x])
		{
		case '+':
			cOps[x] = 0;
			break;
		case '-':
			cOps[x] = 1;
			break;
		case '*':
			cOps[x] = 2;
			break;
		case '/':
			cOps[x] = 3;
			break;
		case '^':
			cOps[x] = 4;
			break;
		case '=':
			cOps[x] = 5;
			break;
		default:
			cout <<"Operator error" <<endl;
			exit(1);
		}
	}
	cP = Postfix::opTable[cOps[0]][cOps[1]];
	return cP;
}

bool Postfix::EvaluatePostfix(G::Queue<char *> &qTokens, float &fResult)
{
	G::Stack<char *> sResult;
	char *szToken;
	float n1 = 0;
	float n2 = 0;
	char *szN1 = NULL;
	char *szN2 = NULL;

//	bool bType;

	while (qTokens.size())
	{
		qTokens.take(szToken);
		// Check type
		switch (szToken[0])
		{
		case '+':
			{
			sResult.pop(szN2);
			sResult.pop(szN1);
			if (!Postfix::GetTokenValue(szN2, n2))
				return false;
			if (!Postfix::GetTokenValue(szN1, n1))
				return false;
			n1 += n2;
			char *szValue = new char[80];
			sprintf(szValue, "%f", n1);
			sResult.push(szValue);
			delete[] szN2;
			delete[] szN1;
			break;
			}
		case '-':
			{
			if (isdigit(szToken[1]))
				goto NUMBER;
			sResult.pop(szN2);
			sResult.pop(szN1);
			if (!Postfix::GetTokenValue(szN2, n2))
				return false;
			if (!Postfix::GetTokenValue(szN1, n1))
				return false;
			n1 -= n2;
			char *szValue = new char[80];
			sprintf(szValue, "%f", n1);
			sResult.push(szValue);
			delete[] szN2;
			delete[] szN1;
			break;
			}
		case '*':
			{
			sResult.pop(szN2);
			sResult.pop(szN1);
			if (!Postfix::GetTokenValue(szN2, n2))
				return false;
			if (!Postfix::GetTokenValue(szN1, n1))
				return false;
			n1 *= n2;
			char *szValue = new char[80];
			sprintf(szValue, "%f", n1);
			sResult.push(szValue);
			delete[] szN2;
			delete[] szN1;
			break;
			}
		case '/':
			{
			sResult.pop(szN2);
			sResult.pop(szN1);
			if (!Postfix::GetTokenValue(szN2, n2))
				return false;
			if (!Postfix::GetTokenValue(szN1, n1))
				return false;
			n1 /= n2;
			char *szValue = new char[80];
			sprintf(szValue, "%f", n1);
			sResult.push(szValue);
			delete[] szN2;
			delete[] szN1;
			break;
			}
		case '^':
			{
			sResult.pop(szN2);
			sResult.pop(szN1);
			if (!Postfix::GetTokenValue(szN2, n2))
				return false;
			if (!Postfix::GetTokenValue(szN1, n1))
				return false;
			n1 = (float)pow((double)n1, (double)n2);
			char *szValue = new char[80];
			sprintf(szValue, "%f", n1);
			sResult.push(szValue);
			delete[] szN2;
			delete[] szN1;
			break;
			}
		case '=':
			sResult.pop(szN2);
			sResult.pop(szN1);
			if (!Postfix::GetTokenValue(szN2, n2))
				return false;
			if (GetTokenType(szN1) != 0)
				return false;
			if (!AddVar(szN1, n2))
				return false;
			sResult.push(szN2);
			delete[] szN1;
			break;
NUMBER:
		default:
			//n1 = (float)atof(szToken);
			sResult.push(szToken);
		}
		n1 = 0;
		n2 = 0;
		szN1 = NULL;
		szN2 = NULL;
	}

	sResult.pop(szToken);
	if (!GetTokenValue(szToken, fResult))
		return false;
	delete[] szToken;
	return true;
}

bool Postfix::CreateTokenList(char **szTokens)
{
	char *szTemp = NULL;
	unsigned int iLen = G::Str::Len(*szTokens);
	G::Str::Copy(&szTemp, *szTokens);
	delete[] *szTokens;
	*szTokens = new char[iLen * 4];
	int cPos = 0;
	//bool bStart = false;
	for (unsigned int i = 0; i < iLen; i++)
	{
		char cCurrent = szTemp[i];

		switch (cCurrent)
		{
		case '+':
		case '-':
		case '*':
		case '/':
		case '^':
		case '(':
		case ')':
		case '=':
			(*szTokens)[cPos] = ' ';
			(*szTokens)[cPos+1] = cCurrent;
			(*szTokens)[cPos+2] = ' ';
			cPos += 3;
			break;
		case ' ':
			break;
		default:
			//if ((cCurrent == '_') && (((*szTokens)[cPos-1] == ' ') || (cPos == 0)))
			//	cCurrent = '-';
			(*szTokens)[cPos] = cCurrent;
			cPos++;
		}
	}
	(*szTokens)[cPos] = '\0';
	return true;
}

bool Postfix::AddVar(const char *szName, float fValue)
{
	using namespace Postfix;
	Var *pVar = NULL;

	if (FindVar(szName, &pVar))
		pVar->fValue = fValue;
	else
	{
		// Create a new var
		if (vVars.size() == iVarCount)
			vVars.redimp(iVarCount * 2);
		vVars[iVarCount].szName = new char[strlen(szName)+1];
		strcpy(vVars[iVarCount].szName, szName);
		vVars[iVarCount].fValue = fValue;
		iVarCount++;
	}
	return true;
}
bool Postfix::AddVar(const char *szName, float fValue, Var **pVar)
{
	using namespace Postfix;
	//Var *pVar = NULL;

	if (FindVar(szName, pVar))
		(*pVar)->fValue = fValue;
	else
	{
		// Create a new var
		if (vVars.size() == iVarCount)
			vVars.redimp(iVarCount * 2);
		vVars[iVarCount].szName = new char[strlen(szName)+1];
		strcpy(vVars[iVarCount].szName, szName);
		vVars[iVarCount].fValue = fValue;
		*pVar = &vVars[iVarCount];
		iVarCount++;
	}
	return true;
}

bool Postfix::FindVar(const char *szName, Var **pVar)
{
	for (unsigned int i = 0; i < iVarCount; i++)
	{
		if (strcmp(szName, vVars[i].szName) == 0)
		{
			*pVar = &vVars[i];
			return true;
		}
	}
	return false;
}

bool Postfix::GetVar(const char *szName, float &fValue)
{
	Var *pVar = NULL;
	if (!FindVar(szName, &pVar))
		return false;
	fValue = pVar->fValue;
	return true;
}

int Postfix::GetTokenType(const char *szToken)
{
	// is numeric?
	int x;
	bool bNumeric = true, bVar = true;
	if (!szToken)
		return -1;
	if (szToken[0] == '-')
		return 1;

	for (x = 0; x < (int)strlen(szToken); x++)
	{
		if (!isdigit(szToken[x]) && (szToken[x] != '.'))
			bNumeric = false;
	}

	if (bNumeric)
		return 1;

	for (x = 0; x < (int)strlen(szToken); x++)
	{
		if (!isalpha(szToken[x]))
			bVar = false;
	}

	if (bVar)
		return 0;
	else
		return -1;
}

bool Postfix::GetTokenValue(const char *szToken, float &fValue)
{
	int bType;

	bType = Postfix::GetTokenType(szToken);

	if (bType == -1)
	{
		cout <<"Token error" <<endl;
		return false;
	}
	else if (bType == 0)
	{
		// its a variable
		if (!Postfix::GetVar(szToken, fValue))
			fValue = 0;
	}
	else
		fValue = (float)atof(szToken);

	return true;
}

bool Postfix::IsWhiteSpace(const char *szText)
{
	bool bWhite = true;
	unsigned int iLen = G::Str::Len(szText);
	for (unsigned int i = 0; i < iLen; i++)
	{
		switch (szText[i])
		{
		case ' ':
		case '\t':
		case '\r':
		case '\n':
			break;
		default:
			bWhite = false;
		}
		if (!bWhite)
			break;
	}
	return bWhite;
}